using System;
using System.Diagnostics;
using System.Threading.Tasks;

// from: https://docs.microsoft.com/de-de/dotnet/standard/parallel-programming/

class MultiplyMatrices
{
    #region Sequential_Loop
    static void MultiplyMatricesSequential(double[,] matA, double[,] matB,
                                           double[,] result)
    {
        int matACols = matA.GetLength(1);
        int matBCols = matB.GetLength(1);
        int matARows = matA.GetLength(0);

        for (int i = 0; i < matARows; i++)
        {
            for (int j = 0; j < matBCols; j++)
            {
                double temp = 0;
                for (int k = 0; k < matACols; k++)
                {
                   temp += matA[i, k] * matB[k, j];
                }
                result[i, j] += temp;
            }
        }
    }
    #endregion

    #region Sequential_Loop
    static void MultiplyMatricesSequentialTranspose(double[,] matA, double[,] matB,
                                                    double[,] result)
    {
        int matACols = matA.GetLength(1);
        int matBCols = matB.GetLength(1);
        int matARows = matA.GetLength(0);
        double[,] matBt = TransposeMatrix(matB);

        for (int i = 0; i < matARows; i++)
        {
            for (int j = 0; j < matBCols; j++)
            {
                double temp = 0;
                for (int k = 0; k < matACols; k++)
                {
                   temp += matA[i, k] * matBt[j, k];
                }
                result[i, j] += temp;
            }
        }
    }
    #endregion



    #region Parallel_Loop
    static void MultiplyMatricesParallel(double[,] matA, double[,] matB, double[,] result)
    {
        int matACols = matA.GetLength(1);
        int matBCols = matB.GetLength(1);
        int matARows = matA.GetLength(0);

        // A basic matrix multiplication.
        // Parallelize the outer loop to partition the source array by rows.
        Parallel.For(0, matARows, i =>
        {
            for (int j = 0; j < matBCols; j++)
            {
                double temp = 0;
                for (int k = 0; k < matACols; k++)
                {
                    temp += matA[i, k] * matB[k, j];
                }
                result[i, j] = temp;
            }
        }); // Parallel.For
    }
    #endregion

    #region Parallel_Loop
    static void MultiplyMatricesParallelTranspose(double[,] matA, double[,] matB, double[,] result)
    {
        int matACols = matA.GetLength(1);
        int matBCols = matB.GetLength(1);
        int matARows = matA.GetLength(0);
        double[,] matBt = TransposeMatrix(matB);

        // A basic matrix multiplication.
        // Parallelize the outer loop to partition the source array by rows.
        Parallel.For(0, matARows, i =>
        {
            for (int j = 0; j < matBCols; j++)
            {
                double temp = 0;
                for (int k = 0; k < matACols; k++)
                {
                    temp += matA[i, k] * matBt[j, k];
                }
                result[i, j] = temp;
            }
        }); // Parallel.For
    }
    #endregion



    #region Main
    static void Main(string[] args)
    {
        // Set up matrices. Use small values to better view 
        // result matrix. Increase the counts to see greater 
        // speedup in the parallel loop vs. the sequential loop.
        int colCount = 1000; //180;
        int rowCount = colCount; //2000;
        int colCount2 = colCount; //270;
        double[,] m1 = InitializeMatrix(rowCount, colCount);
        double[,] m2 = InitializeMatrix(colCount, colCount2);
        double[,] result = new double[rowCount, colCount2];
        double[,] result2 = new double[rowCount, colCount2];

        // First do the sequential version.
        Console.Error.WriteLine("Executing sequential loop...");
        Stopwatch stopwatch = new Stopwatch();

        stopwatch.Start();
        MultiplyMatricesSequential(m1, m2, result);
        stopwatch.Stop();
        Console.Error.WriteLine("Sequential loop time in milliseconds: {0}",
                                stopwatch.ElapsedMilliseconds);
        //OfferToPrint(rowCount, colCount2, result);

        stopwatch.Reset();
        stopwatch.Start();
        MultiplyMatricesSequentialTranspose(m1, m2, result2);
        stopwatch.Stop();
        Console.Error.WriteLine("Sequential transpose loop time in milliseconds: {0}",
                                stopwatch.ElapsedMilliseconds);
        Console.Error.WriteLine("equal sequential matrix: {0}", EqualMatrix(result,result2));


        // Reset timer. 
        stopwatch.Reset();
        // Do the parallel loop.
        Console.Error.WriteLine("Executing parallel loop...");
        stopwatch.Start();
        MultiplyMatricesParallel(m1, m2, result2);
        stopwatch.Stop();
        Console.Error.WriteLine("Parallel loop time in milliseconds: {0}",
                                stopwatch.ElapsedMilliseconds);
        //OfferToPrint(rowCount, colCount2, result);
        Console.Error.WriteLine("equal sequential-parallel matrix: {0}", EqualMatrix(result,result2));

        // Reset timer and results matrix. 
        stopwatch.Reset();
        // Do the parallel loop.
        Console.Error.WriteLine("Executing parallel loop...");
        stopwatch.Start();
        MultiplyMatricesParallelTranspose(m1, m2, result);
        stopwatch.Stop();
        Console.Error.WriteLine("Parallel transpose loop time in milliseconds: {0}",
                                stopwatch.ElapsedMilliseconds);
        Console.Error.WriteLine("equal parallel matrix: {0}", EqualMatrix(result,result2));

    }
    #endregion


    static double[,] InitializeMatrix(int rows, int cols)
    {
        double[,] matrix = new double[rows, cols];
        Random r = new Random();
        for (int i = 0; i < rows; i++)
        {
            for (int j = 0; j < cols; j++)
            {
                matrix[i, j] = r.Next(100);
            }
        }
        return matrix;
    }

    static double[,] TransposeMatrix(double[,] mat)
    {
        int rows = mat.GetLength(0);
        int cols = mat.GetLength(1);
        double[,] matrix = new double[cols, rows];
        for (int i = 0; i < cols; i++)
        {
            for (int j = 0; j < rows; j++)
            {
                matrix[i, j] = mat[j, i];
            }
        }
        return matrix;
    }

    static Boolean EqualMatrix(double[,] matA, double[,] matB)
    {
        int rows = matA.GetLength(0);
        int cols = matA.GetLength(1);
        if (rows != matB.GetLength(0)) 
        {
            return false;
        }
        if (cols != matB.GetLength(1)) 
        {
            return false;
        }

        for (int i = 0; i < cols; i++)
        {
            for (int j = 0; j < rows; j++)
            {
                if (matA[i, j] != matB[i, j])
                {
                    return false;
                }
            }
        }
        return true;
    }

    private static void OfferToPrint(int rowCount, int colCount, double[,] matrix)
    {
        Console.Error.Write("Computation complete. Print results (y/n)? ");
        char c = Console.ReadKey(true).KeyChar;
        Console.Error.WriteLine(c);
        if (Char.ToUpperInvariant(c) == 'Y')
        {
            //if (! Console.IsOutputRedirected) Console.WindowWidth = 180;
            Console.WriteLine();
            for (int x = 0; x < rowCount; x++)
            {
                Console.WriteLine("ROW {0}: ", x);
                for (int y = 0; y < colCount; y++)
                {
                    Console.Write("{0:#.##} ", matrix[x, y]);
                }
                Console.WriteLine();
            }
        }
    }

}
