/*
 * $Id: MatrixMultOmp.jomp,v 1.5 2002/08/03 17:14:31 kredel Exp $
 */

//package edu.unima.ky.parallel.jomp;

/**
 * Matrix Multiplication 
 * sequential and parallel using OpenMP
 * @author Akitoshi Yoshida
 * @author Heinz Kredel.
 */

public class MatrixMultOmp {


/**
 * @param C two-dimensional double array.
 * @param A two-dimensional double array.
 * @param B two-dimensional double array.
 */
  public void seqmult(double[][] C, double[][] A, double[][] B) {
      for (int i=0; i < A.length; i++) {
          for (int j=0; j < B[0].length; j++) {
              double c = 0.0;
              for (int k=0; k < B.length; k++) {
                  c += A[i][k] * B[k][j];
              }
              C[i][j] = c;
          }
      }
  }


/**
 * @param C two-dimensional double array.
 * @param A two-dimensional double array.
 * @param B two-dimensional double array.
 */
  public void parmult(double[][] C, double[][] A, double[][] B) {

// OMP PARALLEL BLOCK BEGINS
{
  __omp_Class0 __omp_Object0 = new __omp_Class0();
  // shared variables
  __omp_Object0.B = B;
  __omp_Object0.A = A;
  __omp_Object0.C = C;
  // firstprivate variables
  try {
    jomp.runtime.OMP.doParallel(__omp_Object0);
  } catch(Throwable __omp_exception) {
    System.err.println("OMP Warning: Illegal thread exception ignored!");
    System.err.println(__omp_exception);
  }
  // reduction variables
  // shared variables
  B = __omp_Object0.B;
  A = __omp_Object0.A;
  C = __omp_Object0.C;
}
// OMP PARALLEL BLOCK ENDS

  }


/**
 * @param C two-dimensional double array.
 * @param A two-dimensional double array.
 * @param B two-dimensional double array.
 */
  public void seqdiff(double[][] C, double[][] A, double[][] B) {
      for (int i=0; i < C.length; i++) {
          for (int j=0; j < C[0].length; j++) {
              C[i][j] = A[i][j] - B[i][j];
          }
      }
  }


/**
 * @param C two-dimensional double array.
 * @param A two-dimensional double array.
 * @param B two-dimensional double array.
 * @param i row of A.
 * @param j column of B.
 */
  public void dotmult(double[][] C, double[][] A, double[][] B, int i, int j) {
      double c = 0.0;
      for (int k=0; k < B.length; k++) {
          c += A[i][k] * B[k][j];
      }
      C[i][j] = c;
  }


/**
 * @param C two-dimensional double array.
 * @param A two-dimensional double array.
 * @param B two-dimensional double array.
 */
  public void seq2mult(double[][] C, double[][] A, double[][] B) {
      for (int i=0; i < A.length; i++) {
          for (int j=0; j < B[0].length; j++) {
              dotmult(C,A,B,i,j);
          }
      }
  }


/**
 * @param n rows of result.
 * @param m columns of result.
 * @return A two-dimensional double array.
 */
  public double[][] matgen(int n, int m) {
      double A[][] = new double[n][m];
      for (int i=0; i < n; i++) {
          for (int j=0; j < m; j++) {
              A[i][j] = Math.random();
          }
      }
      return A;
  }


/**
 * @param n rows of result.
 * @param m columns of result.
 * @return A two-dimensional double array.
 */
  public double[][] matgen0(int n, int m) {
      double A[][] = new double[n][m];
      for (int i=0; i < n; i++) {
          for (int j=0; j < m; j++) {
              A[i][j] = 0.0;
          }
      }
      return A;
  }


/**
 * @param n rows of result.
 * @param m columns of result.
 * @return A two-dimensional double array.
 */
  public double[][] matgen1(int n, int m) {
      double A[][] = new double[n][m];
      for (int i=0; i < n; i++) {
          for (int j=0; j < m; j++) {
              if (i == j) A[i][j] = 1.0; else A[i][j]= 0.0;
          }
      }
      return A;
  }
   

/**
 * @param A two-dimensional double array. 
 */
  public void matprint(double[][] A) {
      for (int i=0; i < A.length; i++) {
          for (int j=0; j < A[0].length; j++) {
              System.out.print(A[i][j] + " ");
          }
          System.out.println();
      }
  }


/**
 * @param A two-dimensional double array.
 * @return true if A is approximately zero.
 */
  public boolean matcheck0(double[][] A) {
      double eps = Double.MIN_VALUE*1000.0;
      for (int i=0; i < A.length; i++) {
          for (int j=0; j < A[0].length; j++) {
              if ( Math.abs(A[i][j]) > eps) return false;
          }
      }
      return true;
  }


/**
 */
  public static void main(String[] args) {

        int n = 200, m = 400;
        boolean prnt = false;

        try { n = Integer.parseInt(args[0]); }
        catch (Exception e) { }
        try { m = Integer.parseInt(args[1]); }
        catch (Exception e) { }
        try { prnt = Boolean.valueOf(args[2]).booleanValue(); }
        catch (Exception e) { }

        MatrixMultOmp x = new MatrixMultOmp();

        double[][] A = x.matgen1(n,m);
        double[][] B = x.matgen(m,n);
        double[][] C = x.matgen0(n,n);
        double[][] D = x.matgen0(n,n);
 
        System.out.println("A = ");
        if (prnt) x.matprint(A);
        System.out.println("");

        System.out.println("B = ");
        if (prnt) x.matprint(B);
        System.out.println("");

        System.out.println("C = ");
        long tm = System.currentTimeMillis();
        x.seqmult(C,A,B);
        tm = System.currentTimeMillis() - tm; 
        System.out.println(tm + "ms");
        //x.seq2mult(C,A,B);
        tm = System.currentTimeMillis();
        x.parmult(C,A,B);
        tm = System.currentTimeMillis() - tm; 
        System.out.println(tm + "ms");
        if (prnt) x.matprint(C);
        System.out.println("");

        System.out.println("D = ");
        x.seqdiff(D,B,C);
        if (prnt) x.matprint(D);
        System.out.println("");
        System.out.println("D is zero = " + x.matcheck0(D) );

    }

// OMP PARALLEL REGION INNER CLASS DEFINITION BEGINS
private class __omp_Class0 extends jomp.runtime.BusyTask {
  // shared variables
  double [ ] [ ] B;
  double [ ] [ ] A;
  double [ ] [ ] C;
  // firstprivate variables
  // variables to hold results of reduction

  public void go(int __omp_me) throws Throwable {
  // firstprivate variables + init
  // private variables
  // reduction variables, init to default
    // OMP USER CODE BEGINS

          { // OMP FOR BLOCK BEGINS
          // copy of firstprivate variables, initialized
          // copy of lastprivate variables
          // variables to hold result of reduction
          boolean amLast=false;
          {
            // firstprivate variables + init
            // [last]private variables
            // reduction variables + init to default
            // -------------------------------------
            jomp.runtime.LoopData __omp_WholeData2 = new jomp.runtime.LoopData();
            jomp.runtime.LoopData __omp_ChunkData1 = new jomp.runtime.LoopData();
            __omp_WholeData2.start = (long)(0);
            __omp_WholeData2.stop = (long)( A.length);
            __omp_WholeData2.step = (long)(1);
            jomp.runtime.OMP.setChunkStatic(__omp_WholeData2);
            while(!__omp_ChunkData1.isLast && jomp.runtime.OMP.getLoopStatic(__omp_me, __omp_WholeData2, __omp_ChunkData1)) {
            for(;;) {
              if(__omp_WholeData2.step > 0) {
                 if(__omp_ChunkData1.stop > __omp_WholeData2.stop) __omp_ChunkData1.stop = __omp_WholeData2.stop;
                 if(__omp_ChunkData1.start >= __omp_WholeData2.stop) break;
              } else {
                 if(__omp_ChunkData1.stop < __omp_WholeData2.stop) __omp_ChunkData1.stop = __omp_WholeData2.stop;
                 if(__omp_ChunkData1.start > __omp_WholeData2.stop) break;
              }
              for(int i = (int)__omp_ChunkData1.start; i < __omp_ChunkData1.stop; i += __omp_ChunkData1.step) {
                // OMP USER CODE BEGINS
 {
          for (int j=0; j < B[0].length; j++) {
              double c = 0.0;
              for (int k=0; k < B.length; k++) {
                  c += A[i][k] * B[k][j];
              }
              C[i][j] = c;
          }
      }
                // OMP USER CODE ENDS
                if (i == (__omp_WholeData2.stop-1)) amLast = true;
              } // of for 
              if(__omp_ChunkData1.startStep == 0)
                break;
              __omp_ChunkData1.start += __omp_ChunkData1.startStep;
              __omp_ChunkData1.stop += __omp_ChunkData1.startStep;
            } // of for(;;)
            } // of while
            // call reducer
            jomp.runtime.OMP.doBarrier(__omp_me);
            // copy lastprivate variables out
            if (amLast) {
            }
          }
          // set global from lastprivate variables
          if (amLast) {
          }
          // set global from reduction variables
          if (jomp.runtime.OMP.getThreadNum(__omp_me) == 0) {
          }
          } // OMP FOR BLOCK ENDS

    // OMP USER CODE ENDS
  // call reducer
  // output to _rd_ copy
  if (jomp.runtime.OMP.getThreadNum(__omp_me) == 0) {
  }
  }
}
// OMP PARALLEL REGION INNER CLASS DEFINITION ENDS

}

