/*
 * $Id: BoundedBuffer.java,v 1.7 2004/01/16 17:19:02 kredel Exp $
 */

//package edu.unima.ky.parallel;

/** 
 * A bounded buffer has some capacity to store objects.
 * With the method <code>put</code> objects are placed into 
 * the buffer, and with the <code>get</code> objects are taken 
 * from the buffer.
 * <code>put</code> blocks if there is no space to 
 * place a object into the buffer.
 * <code>get</code> blocks if there are no objects in the buffer.
 * The objects are put to and get from the buffer in FIFO order.
 * @author Akitoshi Yoshida
 * @author Heinz Kredel. 
 * @see Queue
 * @see Deque
 * @see Channel
 * @see MemoryChannel
 * @see SyncChannel
 * @see Stack
 */ 

public class BoundedBuffer extends Object {

  /** 
    * The buffer storage of objects.
    */ 
  private Object[] buffer;

  /** 
    * The size of the buffer.
    */
  private int size;

  /** 
    * The position of the first filled cell. 
    */ 
 private int front;

  /** 
    * The position of the first empty cell. 
    */ 
  private int rear;

  /** 
    * A semaphore to indicate an empty buffer. 
    */ 
  private Semaphore empty;

  /** 
    * A semaphore to indicate a full buffer.
    */ 
  private Semaphore full;

  private Object pt;
  private Object gt;

  /**
    * Constructs a BoundedBuffer with a desired size. 
    * @param init An integer, the size of the buffer.
    */
  public BoundedBuffer(int init) {
    buffer = new Object[init];
    size = init;
    front = 0;
    rear = 0;
    empty = new Semaphore(init);
    full = new Semaphore(0);
    pt = new Object();
    gt = new Object();
  }


  /**
    * Tests if the BoundedBuffer is empty. 
    * If it is, gives back the value true.
    * @return The boolean value of 
         (whether the BoundedBuffer is empty or not).   
    */
  public boolean empty() {
    return (front == rear);
  }


  /**
   * Add an object to the BoundedBuffer
   * @param v An object.
   */
  public void put(Object v) throws InterruptedException {
    empty.P();
    synchronized (pt) {
      if (rear >= size) rear -= size;
      buffer[ rear ] = v; 
      rear++;     
    }
    full.V();
  }

  /**
   * Get an object from the BoundedBuffer
   */
  public Object get()  throws InterruptedException {
    Object v;
    full.P();
    synchronized (gt) {
      if (front >= size) front -= size;
      v = buffer[ front ]; 
      front++; 
    }
    empty.V();
    return v;
  }
}
