
import java.io.IOException;
import java.io.Serializable;
import java.io.FileNotFoundException;
import java.util.ArrayList;

/**
 * A parallel algorithm for an euclidean 2d TSP Problem
 */

public class DistTSP implements TSPInf {

    protected Graph graph;
    protected Point[] points;
    //protected Path bestPath; // beware of threads
    protected DistBestStore best;

    protected int number = 0;
    protected String hostfile = "maschines";
    protected int port = -1;
    protected String master = "localhost";
    protected ChannelFactory cf;

    private static Logger logger = new Logger(2);

    public DistTSP(Point[] p, int th) {
        this.points = p;
        graph = new PlaneGraph(p);
        number = th;
        // distribute code
        ExecutableChannels ec = null;
        try {
            ec = new ExecutableChannels( hostfile );
            ec.open( number );
            master = ec.getMasterHost();
            port = ec.getMasterPort();
            for ( int i = 0; i < number; i++ ) {
                ec.send(i, new DistParClientTSP(master,port,points) );
            }
            System.out.println("code distributed");
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            return;
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        if ( ec != null ) {
           ec.close();
        }
        cf = new ChannelFactory(port);
    }


    public Path actualBest() {
        return best.getPath();
    }

    public void setBest(Path b) {
        if ( b == null ) {
           return;
        }
        if ( best == null ) {
           // best.setPath( b );
           return;
        }
        if ( b.cost() < best.getPath().cost() ) {
           best.setPath( b );
        }
    }


    public Path getBest() {
        return getBest(Long.MAX_VALUE);
    }

    public Path getBest(long max) {
        Path bestPath = PathByteArray.standardPath(graph); // standard path
        if ( bestPath.length() <= 2 ) {
            return bestPath;
        }
        best = new DistBestStore( bestPath, 0, number );
        Path path = PathByteArray.emptyPath(graph); // empty path
        Deque globalStack = new Deque(10000);
        try {
            globalStack.push( path );
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        Thread[] t = new Thread[number];
        for ( int i = 0; i < number; i++ ) {
            SocketChannel sc = null;
            try {
                sc = cf.getChannel();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            t[i] = new DistRunTSP(globalStack,best,max,sc);
            t[i].start();
        }
        try {
            for ( int i = 0; i < number; i++ ) {
                t[i].join();
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        cf.terminate();
        return best.getPath();
    }

    public void getBest(Path path) {
        System.out.println("method getBest() in DistTSP not implemented");
    }

}

class DistBestStore {

    private Path path;
    private int idlers;
    public final int threads;

    public DistBestStore(Path b, int i, int t) {
        path = b;
        idlers = i;
        threads = t;
    }

    public Path getPath() {
        return path;
    }

    public void setPath(Path p) {
        path = p;
    }

    public synchronized void incIdle() {
        idlers++;
    }

    public synchronized void decIdle() {
        idlers--;
    }

    public synchronized boolean allIdle() {
        return (idlers == threads);
    }
}


class DistRunTSP extends Thread {

    private DistBestStore best;
    private Deque globalStack;
    private long maxiter;
    private long iter;
    private long pushes;
    private ArrayList stack;
    private SocketChannel sc;
    private static Logger logger = new Logger(2);

    public DistRunTSP(Deque s, DistBestStore b, long max, SocketChannel sc) {
        globalStack = s;
        best = b;
        maxiter = max;
        stack = new ArrayList(500);
        this.sc = sc;
    }


    public void run() {
        iter = 0;
        pushes = 0;
        Object o = null;
        Path p = null;
        // boolean running = true;
        while (true) {
            // protocol: receive/send/send/receive
            // protocol: receive/push
            o = null;
            try {
                o = sc.receive();
                iter++;
                logger.debug("requested work "+o);
            } catch (IOException e) {
               e.printStackTrace();
               break;
            } catch (ClassNotFoundException e) {
               e.printStackTrace();
               break;
            }
            if ( ! (o instanceof String) ) {
                break;
            }
            if ( "get".equals( (String)o ) ) {
              try {
                best.incIdle();
                if ( globalStack.empty() && best.allIdle() ) {
                   for ( int i = 0; i < best.threads-1; i++) {
                       globalStack.push(null);
                   }
                   best.decIdle();
                   break; //return;
                }
                o = globalStack.pop();
                best.decIdle();
                if ( o == null ) {
                    sc.send( o );
                    break;
                }
                sc.send( best.getPath() );
                logger.debug("send best = " + best.getPath());
                p = (Path)o;
                sc.send( p );
                logger.debug("send work = " + p );
                o = sc.receive();
                logger.debug("received = " + o );
                if ( o == null ) {
                    break;
                }
                while ( o instanceof String && "part".equals( (String)o ) ) {
                    if ( ! handlePart() ) {
                        break;
                    }
                    o = sc.receive();
                    logger.debug("received = " + o );
                    if ( o == null ) {
                       break;
                    }
                }
                if ( o instanceof Path ) {
                    p = (Path)o;
                    synchronized (best) {
                       if ( p.cost() < best.getPath().cost() ) {
                          best.setPath( p );
                       }
                    }
                } else {
                    break;
                }
              } catch (InterruptedException e) {
                 e.printStackTrace();
                 break;
              } catch (IOException e) {
                 e.printStackTrace();
                 break;
              } catch (ClassNotFoundException e) {
                 e.printStackTrace();
                 break;
              }
            } else if ( "part".equals( (String)o ) ) {
                 if ( ! handlePart() ) {
                      break;
                 }
            } else {
                logger.warn("invalid request = " + o); 
                break;
            }
        } 
        System.out.println("requests = " + iter 
                         + " pushes = " + pushes);
        sc.close();
    }

    protected boolean handlePart() {
         try {
                Object o = sc.receive();
                logger.debug("received part = " + o );
                if ( o == null ) {
                    return false;
                }
                if ( o instanceof Path ) {
                    Path p = (Path)o;
                    pushes++;
                    globalStack.push(p);
                } else {
                    return false;
                }
         } catch (InterruptedException e) {
             e.printStackTrace();
             return false;
         } catch (IOException e) {
             e.printStackTrace();
             return false;
         } catch (ClassNotFoundException e) {
             e.printStackTrace();
             return false;
         }
         return true;
    }

}


/**
 * Objects of this class are to be send to a ExecutableServer
 */

class DistParClientTSP implements RemoteExecutable {

    protected String host;
    protected int port;
    // TSPInf tsp;
    protected Graph graph;
    protected Point[] points;
    protected DistParBestStore best;
    protected int number = 0;
    private ArrayList stack;
    public int count;
    private long iter;
    private long maxiter;
    private int pushes;
    private SocketChannel sc = null;
    private static Logger logger = null;

    public DistParClientTSP(String host, int port, Point[] points) {
	this.host = host;
	this.port = port;
	this.points  = points;
        iter = 0;
        maxiter = Long.MAX_VALUE;
        pushes = 0;
    }

    public Path actualBest() {
        return best.getPath();
    }

    public void setBest(Path b) {
        if ( b == null ) {
           return;
        }
        if ( best == null ) {
           // best.setPath( b );
           return;
        }
        if ( b.cost() < best.getPath().cost() ) {
           best.setPath( b );
        }
    }

    public void run() {
        logger = new Logger(2);
        graph = new PlaneGraph(points);
        Path bestPath = PathByteArray.standardPath(graph); // standard path
        best = new DistParBestStore(bestPath);
        count = best.getPath().maxsize()+5; //*2 / +5 ?
        stack = new ArrayList(500);

        ChannelFactory cf = new ChannelFactory();
        try {
            sc = cf.getChannel( host, port );
        } catch (IOException e) {
            e.printStackTrace();
        }
        System.out.println("DistParClientTSP:" 
                         + " host = " + host
                         + " port = " + port
                         + " best = " + actualBest()
                         + " sc = " + sc
                          );

        Object o;
        Path p;
        int runs = 0;
        while ( true ) {
            try {
                // protocol: send/receive/receive/send 
                logger.debug("request work");
                sc.send( "get" );
                // get best path
                o = sc.receive();
                if ( o == null ) {
                   break;
                }
                if ( o instanceof Path ) {
                   p = (Path)o;
                   if ( p.cost() < actualBest().cost() ) {
                      setBest( p );
                   }
                } else {
                   break;
                }
                // best != null
                // get work
                o = sc.receive();
                if ( o == null ) {
                   break;
                }
                if ( o instanceof Path ) {
                   p = (Path)o;
                   logger.debug("recieved path = "+p);
                   runs++;
                   getParBest(p);
                   // send best
                   sc.send( actualBest() );
                   logger.debug("send best path = "+actualBest());
                } else {
                   break;
                }
	    } catch (IOException ignored) {
                break;
	    } catch (ClassNotFoundException ignored) {
                break;
	    }
        }
        System.out.println("dist client runs = " + runs);
        sc.close();
        cf.terminate();
    }

    public void getParBest(Path path) {
        stack.add( path );
        Object o;
        Path p;
        while ( stack.size() > 0 ) {
            o = stack.remove( stack.size()-1 );
            if ( o == null ) {
                break; //return;
            }
            p = (Path)o;
            getBest(p);
        }
    }

    public void getBest(Path path) {
        Path[] ps = path.nextPaths(); 
        iter += ps.length;
        double bestCost = best.getPath().cost();
        if ( path.length()+1 == path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               if ( ps[i].cost() < bestCost ) {
                 synchronized ( best ) {
                     if ( ps[i].cost() < best.getPath().cost() ) {
                        bestCost = ps[i].cost();
                        best.setPath( (Path)ps[i].clone() ); // evtl. error
                        logger.info("dist par best = " + ps[i]);
                   }
                 }
               }
           }
           return;
        }
        if ( iter >= maxiter ) {
            return;
        }
        if ( path.length()+5 >= path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               if ( ps[i].cost() < bestCost ) { // cut
                   //System.out.print("#");
                   getBestRec( ps[i].copyMax(), path.length()+1 ); // changes best and bestPath
                  //stack.add( ps[i] );
               }
           }
           return;
        }
        for ( int i = 0; i < ps.length; i++ ) { // branch
            if ( ps[i].cost() < bestCost ) { // cut
               // getBest( ps[i] ); // changes best and bestPath
               //System.out.print("-");
               stack.add( ps[i] );
            }
        }
        if ( stack.size() > count && path.length() <= 3 ) {
               Object o = stack.remove( 0 ); //stack.size()-1, depth first
               pushes++;
               try {
                   if ( o != null ) {
                      logger.debug("send work " + o);
                      sc.send( "part" );
                      sc.send( o );
                   }
                   // globalStack.push( o ); // beware of deadlock
               } catch (IOException e) {
                   e.printStackTrace();
               }
        }
        return;
    }

    public void getBestRec(Path path, int depth) {
        double bestCost = best.getPath().cost();
        for ( int k = 0; k < path.maxsize()-depth; k++ ) { // branch
            Path ps = path.nextPath(k); 
            iter++;
            if ( depth+1 == path.maxsize() ) {
               if ( ps.cost() < bestCost ) {
                  synchronized ( best ) {
                     if ( ps.cost() < best.getPath().cost() ) {
                        bestCost = ps.cost();
                        best.setPath( (Path)ps.clone() ); // had error
                        logger.info("dist par best = " + ps);
                     }
                  }
               }
            }
            if ( iter >= maxiter ) {
               return;
            }
            if ( ps.cost() < bestCost ) { // cut
                getBestRec( ps, depth+1 ); // changes best and bestPath
                //stack.add( ps[i] );
            }
        }
        return;
    }

}

class DistParBestStore implements Serializable {

    private Path path;

    public DistParBestStore(Path b) {
        path = b;
    }

    public Path getPath() {
        return path;
    }

    public void setPath(Path p) {
        path = p;
    }
}
